# Cerebrum Medulla

Medulla is an implementation of self modifying cartesian genetic programming, which serves as the dynamic system
in the Cerebrum reservoir computing platform.

*DISCLAIMER:*

*Some parts of the system are remnants of earlier versions and might not work anymore.
It was developed and used one task at the time, adding and removing features as we went.*

## Basic usage

The project uses Maven. Package dependencies are found in the `pom.xml` file. Use `mvn package` to download the
dependencies and create a target JAR file, or let your favourite IDE handle it.

We have enforced GC runs for the memory heavy tasks, but give the application some extra
RAM, as it is likes RAM. RAM is good.

To start the system with 4G of heap space, run the following command:

```
java -Xmx4g -jar target/no.ntnu.cerebrum.medulla-jar-with-dependencies.jar
```

On a fresh run of the system, log and configuration folders will be created from the defaults.
Logs will be put in a `logs` folder in the current working directory where the system is
launched from.

Configuration files will be placed in a `config` folder in the current working directory
where the system is launched from.

Serialized SMCGP phenotypes with associated data will be serialized in a `storage`
folder in the current working directory where the system is launched from.
JSON-serialized files will be stored in `storage/candidates`. If CSV dumping of
raw output is enabled, these output and target value dumps will go into `storage/csv`.

## Supporting scripts

There are also some python scripts in the `scripts` directory that help inspecting the different candidate files.
Individual plots can be made using `gen_graphs.py`. The trusty `inspector.py` is the most feature rich, and can
do several operations to sort on different stats, filter candidates, print various properties and move, copy or delete.
For usage information run `python scripts/inspector.py --help`. Standard usage is `python scripts/inspector.py storage/candidates -s -v -st` which will sort candidates by readout accuracy, print verbose output and add
a final line of statistics with mean, median and stdev for the readout accuracies.

## Configuring the system

The following configuration options are available. Not everything does something that is important.

```
# Main system options

# Task
no.ntnu.cerebrum.medulla.task:
		"generateNew" will evolve new SMCGP programs based on current settings
		"rerun" will re-evaluate candidates in the storage/candidates folder on the current dataset
		"narma" will perform NARMA10 eval on candidates in storage/candidates
		"matrix" will perform temporal parity 3, 5, 7 and density 3, 5, 7 on candidates in storage/candidates
no.ntnu.cerebrum.medulla.task.matrix.overWriteExisting: true if you want to overwrite existing matrix evaluation results
no.ntnu.cerebrum.medulla.task.narma.overWriteExisting= true if you want to overwrite existing narma10 results
no.ntnu.cerebrum.medulla.narma.numRunsPerCandidate= integer, how many NARMA10 runs to perform

# Data set
## General
no.ntnu.cerebrum.medulla.dataset.type=
		temporalParity, temporalDensity or narma10
no.ntnu.cerebrum.medulla.dataset.normalize=
		true if data sets should be scaled to within -1, 1 before training readout
no.ntnu.cerebrum.medulla.dataset.dumpToCSV=
		true if raw output and target labes should be dumped to storage/csv
no.ntnu.cerebrum.medulla.dataset.trainingSamples=
		the number of samples to create (ignored for narma as they are premade)
no.ntnu.cerebrum.medulla.dataset.testSamples=
		the number from the training samples that will be removed and used for test instead
no.ntnu.cerebrum.medulla.dataset.numSamplesPerBatch=
		the number of samples in a batch (will affect the number of batches in a set)

## Temporal Parity/Density specific
no.ntnu.cerebrum.medulla.dataset.temporal.delay=
		the number of time steps in signal delay for temporal parity and density
no.ntnu.cerebrum.medulla.dataset.temporal.windowSize=
		the number of time steps in window size for temporla parity and density

## Evaluation specific
no.ntnu.cerebrum.medulla.dataset.evaluation.spectrum.minPowerCoefficient=
		minimum signal power to be considered in harmonic check
no.ntnu.cerebrum.medulla.dataset.evaluation.spectrum.maxPowerCoefficient=
		maximum signal power to be considered in harmonic check

# SMCGP
## Static
no.ntnu.cerebrum.medulla.smcgp.functionSM=no.ntnu.cerebrum.medulla.smcgp.graph.action.selfmodifying. # Dont change
no.ntnu.cerebrum.medulla.smcgp.functionMath=no.ntnu.cerebrum.medulla.smcgp.graph.action.mathematical.	# Don't change
no.ntnu.cerebrum.medulla.smcgp.functionMath.whiteList=
		This can be "all" or a comma separated list of class names from the grap.action.mathematical folder

## Config
no.ntnu.cerebrum.medulla.smcgp.realnumbersLower=-40.0	# Don't change
no.ntnu.cerebrum.medulla.smcgp.realnumbersUpper=40.0	# Don't change
no.ntnu.cerebrum.medulla.smcgp.todoSize=1
		Controls the number of self modification ops per iteration. Hasn't been tested with more than 1
no.ntnu.cerebrum.medulla.smcgp.startNodes=
		Sets the starting size of the SMCGP graph
no.ntnu.cerebrum.medulla.smcgp.levelsback=
		Sets the maximum connection length for each node in the SMCGP graph
no.ntnu.cerebrum.medulla.smcgp.maxNodesMovedOrDuplicated=
		Sets a maximum size of nodes that can be operated upon in one iteration
no.ntnu.cerebrum.medulla.smcgp.outputs=
		Sets the number of output nodes
no.ntnu.cerebrum.medulla.smcgp.timeout=500 # Increase this if need be, but can greatly reduce performance
no.ntnu.cerebrum.medulla.smcgp.cache=false # No effect, not in use anymore
no.ntnu.cerebrum.medulla.smcgp.maxGrowthFactor=
		Real valued number that controls how large a phenotype can grow
no.ntnu.cerebrum.medulla.smcgp.useSimpleInputFunction=
		If set to true, this will only allow regular input functions that increment the pointer

# Phenotype specific viability check
no.ntnu.cerebrum.medulla.smcgp.phenotype.minFitness=
		Minimum readout layer accuracy for which candidates are saved to disk
no.ntnu.cerebrum.medulla.smcgp.phenotype.minPowerSpectrumStates=
		The minimum number of harmonics for a phenotype to pass fitness eval
no.ntnu.cerebrum.medulla.smcgp.phenotype.powerSpectrumRoundingThreshold=
		Signal power below this threshold will be considered zero
no.ntnu.cerebrum.medulla.smcgp.phenotype.useNumericalConvergenceCheck=
		true if using the numerical convergence fitness check
no.ntnu.cerebrum.medulla.smcgp.phenotype.pointAttractorStepsBack=
		The number of time steps back to look for state changes
no.ntnu.cerebrum.medulla.smcgp.phenotype.numericalConvergenceStepsBack=
		The number of time steps back to look for numerical convergence
no.ntnu.cerebrum.medulla.smcgp.phenotype.numericalConvergenceRoundingThreshold=
		Values within this threshold are considered equal
no.ntnu.cerebrum.medulla.smcgp.phenotype.minUniqueValues=
		Minimum number of unique values required to pass fitness check
no.ntnu.cerebrum.medulla.smcgp.phenotype.minTopographicStates=
		Minimum number of topological states allowed to pass fitness check
no.ntnu.cerebrum.medulla.smcgp.phenotype.maxTopographicStates=
		Maximum number of topological states allowed to pass fitness check
no.ntnu.cerebrum.medulla.smcgp.phenotype.minInputNodePercentage=
		Minimum percentage of nodes in the graph that are of the input type

## EA
no.ntnu.cerebrum.medulla.smcgp.ea.maxEpochs=
		Controls how many epochs before the EA search returns current best
no.ntnu.cerebrum.medulla.smcgp.ea.stopCondition=64	# Don't change!
no.ntnu.cerebrum.medulla.smcgp.ea.mutationRate=
		Controls mutation rate in the EA search
no.ntnu.cerebrum.medulla.smcgp.ea.initialPopulation=
		Sets the starting population size
no.ntnu.cerebrum.medulla.smcgp.ea.poolSize=
		Sets the population pool size for the regular population
### Parent Selection Scheme & Breeding
no.ntnu.cerebrum.medulla.smcgp.ea.crossover=
		true if using crossover
no.ntnu.cerebrum.medulla.smcgp.ea.crossoverRate=
		Control the rate for which crossover occurs
no.ntnu.cerebrum.medulla.smcgp.ea.elitism=
		true if using elitism
no.ntnu.cerebrum.medulla.smcgp.ea.parentPool=
		Sets the size of the parent pool used to create offspring
no.ntnu.cerebrum.medulla.smcgp.ea.parentSelectionScheme=
		tournament (best from random) || best (parentPool best in population) || 1+X (one best + X children)
no.ntnu.cerebrum.medulla.smcgp.ea.tournamentSize=
		sets the number of individuals drawn in each tournament

# Signal processing
no.ntnu.cerebrum.medulla.smcgp.signal.seconds=
		Sets the period for which sine wave signal is captured
no.ntnu.cerebrum.medulla.smcgp.signal.amplitude=
		Sets the amplitude of the generated signal
no.ntnu.cerebrum.medulla.smcgp.signal.phase=
		Sets the phase shift of the generated signal
no.ntnu.cerebrum.medulla.smcgp.signal.frequency=
		Sets the frequency of the generated signal
no.ntnu.cerebrum.medulla.smcgp.signal.overSamplingFactor=
		Sets the over sampling factor, i.e the signal resolution

# Reservoir
no.ntnu.cerebrum.medulla.reservoir.epochs=
		Sets the max number of epochs for training the readout layer
no.ntnu.cerebrum.medulla.reservoir.evaluateEveryNumEpochs=
		Evaluate readout layer accuracy every N epochs, printing stats to logs
no.ntnu.cerebrum.medulla.reservoir.numCachedTimeSteps=
		Sets the number of time steps to feed back into the reservoir
no.ntnu.cerebrum.medulla.reservoir.useInputExpansion=
		true if using input expansion
no.ntnu.cerebrum.medulla.reservoir.inputExpansionFactor=
		Sets how many times single input values should be duplicated
no.ntnu.cerebrum.medulla.reservoir.readout.useMultilayerForRegression=
		true if using multi layer regression
no.ntnu.cerebrum.medulla.reservoir.readout.stopCondition=
		Stop training if accuracy reaches this threshold
no.ntnu.cerebrum.medulla.reservoir.readout.batchSize=
		Sets the number of samples in each batch when training the readout layer
no.ntnu.cerebrum.medulla.reservoir.readout.hiddenLayerOneNodeCount=
		Sets the number of nodes in the hidden layer
no.ntnu.cerebrum.medulla.reservoir.readout.learningRate=
		Sets the learning rate of the readout layer
no.ntnu.cerebrum.medulla.reservoir.readout.momentum=
		Sets the momentum factor in Nesterov's Accelerated Gradient
no.ntnu.cerebrum.medulla.reservoir.readout.quarterWayFitnessCutoff=
		Stop training if readout layer accuracy is less than this 1/4 of max epochs
no.ntnu.cerebrum.medulla.reservoir.readout.halfWayFitnessCutoff=
		Stop training if readout layer accuracy is less than this 1/2 of max epochs
no.ntnu.cerebrum.medulla.reservoir.readout.useFitnessTrainingCutoff=
		true if you want to use the above early training stop thresholds
```

