package no.ntnu.cerebrum.medulla.common;

import java.util.List;

/**
 * Created on 09.04.2017.
 */
public class TreeNode {

    private String name;
    private final List<TreeNode> children;

    public TreeNode(String name, List<TreeNode> children) {
        this.name = name;
        this.children = children;
    }

    /**
     * Creates a string representation of the current tree structure, in the same format as the UNIX command line
     * tool "tree", using ASCII-characters to draw connections. The tree is drawn from left to right, with increasing
     * node depth.
     * @return A UNIX "tree" like string.
     */
    public String toString() {
        return toString("", true).trim();
    }

    /**
     * Recursively build the string representation of the tree structure.
     * @param prefix Current line prefix.
     * @param isTail True if this is a tail node, false otherwise
     * @return A UNIX-like tree string.
     */
    private String toString(String prefix, boolean isTail) {
        StringBuilder output = new StringBuilder();

        output.append(prefix);
        output.append(isTail ? "└── " : "├── ");
        output.append(name);
        output.append("\n");

        for (int i = 0; i < children.size() - 1; i++) {
            output.append(
                children.get(i).toString(prefix + (isTail ? "    " : "│   "), false)
            );
        }

        if (children.size() > 0) {
            output.append(
                children.get(children.size() - 1).toString(prefix + (isTail ? "    " : "│   "), true)
            );
        }

        return output.toString();
    }
}
