package no.ntnu.cerebrum.medulla.data;

import java.util.Arrays;

/**
 * Created 2017-03-20.
 *
 * cerebrum-medulla is licenced under the MIT licence.
 */
public class Data {
    private double[][] input;
    private double[][] output;
    private double[] flatInput;
    private double[] flatOutput;

    public Data(int numSamples) {
        input = new double[numSamples][];
        output = new double[numSamples][];
    }

    public int numSamples() {
        return input.length;
    }

    public double[][] input() {
        return input;
    }

    /**
     * Flattens the contents of the input matrix into a 1D array, and caches it for later queries.
     * @return A 1D array representation of the input matrix.
     */
    public double[] flatInput() {
        if (flatInput == null || flatInput.length != input.length * input[0].length) flatInput = DataFactory.flattenArray(input);
        return flatInput;
    }

    public double[][] output() {
        return output;
    }

    /**
     * Flattens the contents of the output matrix into a 1D array, and caches it for later queries.
     * @return A 1D array representation of the output matrix.
     */
    public double[] flatOutput() {
        if (flatOutput == null || flatOutput.length != output.length * output[0].length) flatOutput = DataFactory.flattenArray(output);
        return flatOutput;
    }

    public String toString() {
        String out = "Input:\n[";
        for (int i = 0; i < input.length; i++) {
            if (i == 0) out += Arrays.toString(input[i]) + ",\n";
            else if (i == input.length - 1) out += " " + Arrays.toString(input[i]);
            else out += " " + Arrays.toString(input[i]) + ",\n";
        }
        out += "]\nTarget:\n[";
        for (int i = 0; i < output.length; i++) {
            if (i == 0) out += Arrays.toString(output[i]) + ",\n";
            else if (i == output.length - 1) out += " " + Arrays.toString(output[i]);
            else out += " " + Arrays.toString(output[i]) + ",\n";
        }
        return out + "]";
    }
}
