package no.ntnu.cerebrum.medulla.reservoir.readout;

import no.ntnu.cerebrum.medulla.config.Config;
import org.deeplearning4j.eval.Evaluation;
import org.deeplearning4j.nn.api.OptimizationAlgorithm;
import org.deeplearning4j.nn.conf.MultiLayerConfiguration;
import org.deeplearning4j.nn.conf.NeuralNetConfiguration;
import org.deeplearning4j.nn.conf.Updater;
import org.deeplearning4j.nn.conf.layers.DenseLayer;
import org.deeplearning4j.nn.conf.layers.OutputLayer;
import org.deeplearning4j.nn.multilayer.MultiLayerNetwork;
import org.deeplearning4j.nn.weights.WeightInit;
import org.nd4j.linalg.activations.Activation;
import org.nd4j.linalg.dataset.api.DataSet;
import org.nd4j.linalg.dataset.api.iterator.DataSetIterator;
import org.nd4j.linalg.lossfunctions.LossFunctions;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Created 2017-03-10.
 * <p>
 * cerebrum-medulla is licenced under the MIT licence.
 */
public class MultiLayerClassifier extends ReadoutLayer {
    private Logger log = LoggerFactory.getLogger(MultiLayerClassifier.class);

    public MultiLayerClassifier(int nIn, int nOut) {
        super(nIn, nOut);
    }

    @Override
    public void init(double learningRate, double momentum) {
        log.debug("Initializing multi layer classification network...");

        int hiddenLayerOne = Config.getInt("no.ntnu.cerebrum.medulla.reservoir.readout.hiddenLayerOneNodeCount");

        MultiLayerConfiguration conf = new NeuralNetConfiguration.Builder()
            // Use stochastic gradient descent as an optimization algorithm
            .optimizationAlgo(OptimizationAlgorithm.STOCHASTIC_GRADIENT_DESCENT)
            .updater(Updater.NESTEROVS)
            // Apply updates for each iteration
            .iterations(1)
            // Specify learning rate and the rate of change of the learning rate.
            .learningRate(learningRate)
            .momentum(momentum)
            .biasInit(0.5)
            .list()
            .layer(0, new DenseLayer.Builder()
                .nIn(nIn)
                .nOut(hiddenLayerOne)
                .weightInit(WeightInit.RELU_UNIFORM)
                .activation(Activation.RELU)
                .build())
            // Create output layer
            .layer(1, new OutputLayer.Builder(LossFunctions.LossFunction.MCXENT)
                .nIn(hiddenLayerOne)
                .nOut(nOut)
                .activation(Activation.SOFTMAX)
                .build())
            // No pre train
            .pretrain(false).backprop(true)
            .build();

        net = new MultiLayerNetwork(conf);
        net.init();

        log.debug("Readout layer initialized: " + this);
    }

    @Override
    public double test(DataSetIterator testSet) {
        Evaluation e = new Evaluation();
        while (testSet.hasNext()) {
            DataSet ds = testSet.next();
            e.eval(ds.getLabels(), net.output(ds.getFeatures()));
        }

        return e.accuracy();
    }
}
