package no.ntnu.cerebrum.medulla.reservoir.readout;

import no.ntnu.cerebrum.medulla.config.Config;
import org.deeplearning4j.nn.api.OptimizationAlgorithm;
import org.deeplearning4j.nn.conf.MultiLayerConfiguration;
import org.deeplearning4j.nn.conf.NeuralNetConfiguration;
import org.deeplearning4j.nn.conf.Updater;
import org.deeplearning4j.nn.conf.layers.DenseLayer;
import org.deeplearning4j.nn.conf.layers.OutputLayer;
import org.deeplearning4j.nn.multilayer.MultiLayerNetwork;
import org.deeplearning4j.nn.weights.WeightInit;
import org.nd4j.linalg.activations.Activation;
import org.nd4j.linalg.lossfunctions.LossFunctions;
import org.slf4j.LoggerFactory;

/**
 * Created 2017-05-22.
 * <p>
 * cerebrum-medulla is licenced under the MIT licence.
 */
public class MultiLayerRegression extends SingleLayerLinearRegression {
    public MultiLayerRegression(int nIn, int nOut) {
        super(nIn, nOut);
        log = LoggerFactory.getLogger(MultiLayerRegression.class);
    }

    @Override
    public void init(double learningRate, double momentum) {
        int hiddenLayerOne = Config.getInt("no.ntnu.cerebrum.medulla.reservoir.readout.hiddenLayerOneNodeCount");

        log.debug("Initializing single layer regression readout layer...");
        MultiLayerConfiguration conf = new NeuralNetConfiguration.Builder()
            // Use stochastic gradient descent as an optimization algorithm
            .optimizationAlgo(OptimizationAlgorithm.STOCHASTIC_GRADIENT_DESCENT)
            // Apply updates for each iteration
            .iterations(1)
            // Specify learning rate and the rate of change of the learning rate.
            .learningRate(learningRate)
            .updater(Updater.NESTEROVS)
            .momentum(momentum)
            .biasInit(0.5)
            .list()
            .layer(0, new DenseLayer.Builder()
                .nIn(nIn)
                .nOut(hiddenLayerOne)
                .activation(Activation.SIGMOID)
                .weightInit(WeightInit.SIGMOID_UNIFORM)
                .build())
            // Create output layer
            .layer(1, new OutputLayer.Builder(LossFunctions.LossFunction.MSE)
                .nIn(hiddenLayerOne)
                .nOut(nOut)
                .activation(Activation.IDENTITY)
                .build())
            // No pre train
            .pretrain(false).backprop(true)
            .build();

        net = new MultiLayerNetwork(conf);
        net.init();

        log.debug("Readout layer initialized: " + this);
    }
}
