package no.ntnu.cerebrum.medulla.reservoir.readout;

import no.ntnu.cerebrum.medulla.config.Config;
import org.deeplearning4j.eval.Evaluation;
import org.deeplearning4j.nn.multilayer.MultiLayerNetwork;
import org.nd4j.linalg.dataset.api.DataSet;
import org.nd4j.linalg.dataset.api.iterator.DataSetIterator;

import java.util.ArrayList;
import java.util.List;

/**
 * Created 2017-02-16.
 * <p>
 * cerebrum-medulla is licenced under the MIT licence.
 */
public abstract class ReadoutLayer {
    public MultiLayerNetwork net;

    public abstract void init(double learningRate, double momentum);
    public abstract double test(DataSetIterator testSet);

    protected static int numericError = 0;
    protected static ArrayList<String> exceptions = new ArrayList<>();

    int nIn, nOut;

    ReadoutLayer(int nIn, int nOut) {
        this.nIn = nIn;
        this.nOut = nOut;
    }

    public void train(DataSetIterator trainingSet) {
        while (trainingSet.hasNext()) {
            DataSet ds = trainingSet.next();
            try {
                net.fit(ds);
            } catch (IllegalArgumentException e) {
                exceptions.add(e.getMessage());
                System.err.println(e.getMessage());
            }
        }
    }

    public void train (DataSet trainingSet) {
        List<org.nd4j.linalg.dataset.DataSet> miniBatches = trainingSet.batchBy(
                Config.getInt("no.ntnu.cerebrum.medulla.reservoir.readout.batchSize")
        );
        try {
            for (DataSet ds : miniBatches) net.fit(ds);
        } catch (IllegalArgumentException e) {
            exceptions.add(e.getMessage());
            System.err.println(e.getMessage());
        }
    }

    public Score test(DataSet testSet) {
        Evaluation e = new Evaluation();
        e.eval(testSet.getLabels(), net.output(testSet.getFeatures()));

        Score s = new Score();
        s.accuracy = e.accuracy();
        s.f1 = e.f1();
        s.precision = e.precision();
        s.recall = e.recall();

        return s;
    }

    /**
     * Retrieve the instance of MultiLayerNetwork currently associated with this ReadoutLayer.
     * @return The currently configured neural network acting as the readout layer.
     */
    public MultiLayerNetwork getNetwork() {
        return net;
    }

    public static String stats() {
        return "numeric errors: " + numericError + ", exceptions: " + exceptions.size();
    }
}
