package no.ntnu.cerebrum.medulla.serializers;

import com.google.gson.*;
import javax.annotation.Nullable;
import no.ntnu.cerebrum.medulla.smcgp.Genotype2;

import java.lang.reflect.Type;

public class GenotypeSerializer implements JsonSerializer<Genotype2>, JsonDeserializer<Genotype2> {
    @Override
    public JsonElement serialize(Genotype2 genotype, Type type, JsonSerializationContext context) {
        Gson gson = new GsonBuilder().create();

        JsonObject jsonGenotype = new JsonObject();

        // Genes
        jsonGenotype.add("genes", gson.toJsonTree(genotype.copyGenes()).getAsJsonArray());

        //uuid
        jsonGenotype.addProperty("uuid", genotype.getUuid());

        return jsonGenotype;
    }

    @Override
    public Genotype2 deserialize(JsonElement json, Type typeOfT, JsonDeserializationContext context) throws JsonParseException {
        Gson gson = new GsonBuilder().create();
        JsonObject jsonGenotype = json.getAsJsonObject();

        if (jsonGenotype.has("genes") && jsonGenotype.has("uuid")) {
            // Genes
            String jsonGenes = jsonGenotype.get("genes").toString();
            double[][] genes = gson.fromJson(jsonGenes, double[][].class);

            // Uuid
            String uuid = gson.fromJson(jsonGenotype.get("uuid").getAsString(), String.class);

            Genotype2 deserialized = new Genotype2(genes);
            deserialized.setUuid(uuid);

            return deserialized;
        } else {
            throw new JsonParseException("Cannot deserialize the given json, missing [genes:(double[][])] or [uuid:(String)]");
        }
    }

    public static JsonObject serializeGenotype(Genotype2 genotype2, @Nullable double[] results) {
        Gson gson = new GsonBuilder()
                .registerTypeHierarchyAdapter(Genotype2.class, new GenotypeSerializer())
                .setPrettyPrinting()
                .create();

        JsonObject serializedGenotype = gson.toJsonTree(genotype2).getAsJsonObject();

        // Append last results to json if present
        if (results != null) {
            // Genes
            serializedGenotype.add("results", gson.toJsonTree(results).getAsJsonArray());
        }

        return serializedGenotype;
    }

    public static Genotype2 deserializeGenotype(String jsonGenotype) {
        Gson gson = new GsonBuilder()
                .registerTypeHierarchyAdapter(Genotype2.class, new GenotypeSerializer())
                .setPrettyPrinting()
                .create();

        return gson.fromJson(jsonGenotype, Genotype2.class);
    }
}
