package no.ntnu.cerebrum.medulla.smcgp;

import no.ntnu.cerebrum.medulla.config.Config;
import no.ntnu.cerebrum.medulla.smcgp.graph.action.FunctionLookup;
import no.ntnu.cerebrum.medulla.smcgp.utility.NumberGenerator;
import org.nd4j.linalg.api.rng.Random;
import org.nd4j.linalg.factory.Nd4j;
import org.slf4j.Logger;

import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;
import java.util.UUID;

public class Genotype2 {

    // Logging
    protected static Logger log = org.slf4j.LoggerFactory.getLogger(Genotype2.class);

    private double[][] genes;
    private String uuid;

    public Genotype2() {
        generateRandomGenes(Config.getInt("no.ntnu.cerebrum.medulla.smcgp.startNodes"));
        uuid = UUID.randomUUID().toString();
    }

    public Genotype2(double[][] geneList) {
        genes = geneList;
        uuid = UUID.randomUUID().toString();
    }

    private void generateRandomGenes(int startNodes) {
        genes = new double[startNodes][];

        for (int geneIndex = 0; geneIndex < startNodes; geneIndex++) {
            genes[geneIndex] = generateRandomGene();
        }
    }

    public static double[] generateRandomGene() {
        double[] gene = new double[7];

        // function [1]
        gene[0] = FunctionLookup.getRandomFunction();

        // connections [2]
        gene[1] = NumberGenerator.generateRandomInt(1, Config.getInt("no.ntnu.cerebrum.medulla.smcgp.levelsback"));
        gene[2] = NumberGenerator.generateRandomInt(1, Config.getInt("no.ntnu.cerebrum.medulla.smcgp.levelsback"));

        // real numbers
        gene[3] = NumberGenerator.generateRandomDouble(Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.realnumbersLower"), Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.realnumbersUpper"));
        gene[4] = NumberGenerator.generateRandomDouble(Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.realnumbersLower"), Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.realnumbersUpper"));
        gene[5] = NumberGenerator.generateRandomDouble(Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.realnumbersLower"), Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.realnumbersUpper"));

        // output flag
        gene[6] = NumberGenerator.generateRandomInt(0, 1);
        return gene;
    }

    public double[][] copyGenes() {
        double[][] copy = new double[genes.length][];
        for (int i = 0; i < genes.length; i++) {
            double[] geneCopy = new double[genes[i].length];
            for (int k = 0; k < genes[i].length; k++) {
                geneCopy[k] = genes[i][k];
            }
            copy[i] = geneCopy;
        }
        return copy;
    }

    public void sureMutation(int noGenesToMutate, boolean forceChange) {
        Set mutated = new HashSet<Integer>();
        while (mutated.size() < noGenesToMutate) {
            int geneToMutate = NumberGenerator.generateRandomInt(0, genes.length - 1);
            int setSize = mutated.size();
            mutated.add(geneToMutate);

            if (setSize < mutated.size()) {
                genes[geneToMutate] = mutateChromosome(genes[geneToMutate], forceChange);
            }
        }
    }

    public void mutate(double probability, boolean forceChange) {
        Random rng = Nd4j.getRandom();
        for (int i = 0; i < genes.length; i++) {
            if (rng.nextDouble() < probability) {
                genes[i] = mutateChromosome(genes[i], forceChange);
            }
        }
    }

    private double[] mutateChromosome(double[] gene, boolean forceChange) {
        int chromosomeToMutate = NumberGenerator.generateRandomInt(0, gene.length - 1);
        double currentChromosome = gene[chromosomeToMutate];

        // Mutate again in case nothing is altered
        while (currentChromosome == gene[chromosomeToMutate] && forceChange) {
            switch (chromosomeToMutate) {
                case 0:
                    gene[chromosomeToMutate] = FunctionLookup.getRandomFunction();
                    break;
                case 1:
                case 2:
                    gene[chromosomeToMutate] = NumberGenerator.generateRandomInt(1, Config.getInt("no.ntnu.cerebrum.medulla.smcgp.levelsback"));
                    break;
                case 3:
                case 4:
                case 5:
                    gene[chromosomeToMutate] = NumberGenerator.generateRandomDouble(Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.realnumbersLower"), Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.realnumbersUpper"));
                    break;
                case 6:
                    gene[chromosomeToMutate] = NumberGenerator.generateRandomInt(0, 1);
                    break;
                default:
                    System.out.println("THIS SHOULD NEVER HAPPEN, you just had to screw it up, didn't you!");
            }

            if (gene[chromosomeToMutate] == currentChromosome && forceChange) {
                log.debug(String.format("Mutation: Unaltered [%s] -> [%s]. Running again.", currentChromosome, gene[chromosomeToMutate]));
            }
        }

        return gene;
    }

    public void printGraph() {
        for (double[] gene : this.genes) {
            System.out.println(Arrays.toString(gene));
        }
    }

    public static Genotype2 crossover(Genotype2 mother, Genotype2 father) {
        // If crossover, breed mother and father
        if (Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.ea.crossoverRate") >= NumberGenerator.generateRandomDouble()) {
            double[][] motherGenes = mother.copyGenes();
            double[][] fatherGenes = father.copyGenes();
            double[][] crossoverGenes = mother.copyGenes();
            int crossoverIndex = NumberGenerator.generateRandomInt(0, motherGenes.length - 1);


            for (int i = 0; i < motherGenes.length; i++) {
                if (i < crossoverIndex) {
                    crossoverGenes[i] = Arrays.copyOf(fatherGenes[i], fatherGenes[i].length);
                }
            }
            return new Genotype2(crossoverGenes);
        } else {
            // Pass through mothers genes
            return new Genotype2(mother.copyGenes());
        }

    }

    public String getUuid() {
        return uuid;
    }

    public void setUuid(String uuid) {
        this.uuid = uuid;
    }
}
