package no.ntnu.cerebrum.medulla.smcgp.evolution;

import no.ntnu.cerebrum.medulla.config.Config;
import no.ntnu.cerebrum.medulla.smcgp.Phenotype2;
import org.nd4j.linalg.api.rng.Random;
import org.nd4j.linalg.factory.Nd4j;

import java.util.ArrayList;
import java.util.List;

/**
 * Created 2017-03-14.
 * <p>
 * cerebrum-medulla is licenced under the MIT licence.
 */
public class Selection {

    /* ADULT SELECTION */

    /**
     * Select adults for next generation based on mixing fit subjects from grown up children, and previous adults.
     * @param adults The adults from the previous generation.
     * @param children The newly grown up individuals.
     * @return An ArrayList of individuals that make it to the next generation
     */
    public static List<Phenotype2> generationalMixing(ArrayList<Phenotype2> adults, ArrayList<Phenotype2> children) {
        int maxPoolSize = Config.getInt("smcgp.ea.maxPoolSize");

        ArrayList<Phenotype2> nextGeneration = new ArrayList<>();

        nextGeneration.addAll(adults);
        nextGeneration.addAll(children);
        nextGeneration.sort(Evolve.cmp);

        return nextGeneration.subList(0, maxPoolSize);
    }

    /* PARENT SELECTION */
    public static ArrayList<Phenotype2> tournamentSelection(ArrayList<Phenotype2> population) {
        Random rng = Nd4j.getRandom();

        int adultPoolSize = Config.getInt("smcgp.ea.adultPoolSize");
        int k = Config.getInt("smcgp.ea.tournamentSelection.K");
        double epsilon = Config.getDouble("smcgp.ea.tournamentSelection.epsilon");

        ArrayList<Phenotype2> parents = new ArrayList<>();
        ArrayList<Phenotype2> tournament = new ArrayList<>();

        while (parents.size() < adultPoolSize) {
            // Prepare for new tournament
            tournament.clear();

            // Select K random individuals from the population
            for (int i = 0; i < k; i++) tournament.add(population.get(rng.nextInt(population.size())));
            // If we are within the threshold, pick the best individual from the
            if (rng.nextDouble() < 1.0 - epsilon) {
                tournament.sort(Evolve.cmp);
                parents.add(tournament.get(0));
            } else {
                parents.add(tournament.get(rng.nextInt()));
            }
        }

        return parents;
    }
}
