package no.ntnu.cerebrum.medulla.smcgp.factories;

import no.ntnu.cerebrum.medulla.config.Config;
import no.ntnu.cerebrum.medulla.data.Candidate;
import no.ntnu.cerebrum.medulla.smcgp.Genotype2;
import no.ntnu.cerebrum.medulla.smcgp.Phenotype2;
import no.ntnu.cerebrum.medulla.smcgp.graph.action.FunctionLookup;

import java.util.Arrays;

public class SMCGPFactory {

    private static Genotype2 current;

    public static Genotype2 connectionFloppingGenotype() {
        int genoSize = Config.getInt("no.ntnu.cerebrum.medulla.smcgp.startNodes");
        double[][] genes = new double[genoSize][];

        // #1 - Mean (output)
        double[] mean = Genotype2.generateRandomGene();
        mean[0] = FunctionLookup.getIndexForFunction("Mean");
        mean[1] = 2.0;
        mean[2] = 1.0;
        mean[6] = 1.0; // (output)
        genes[genoSize - 1] = mean;

        // #1 - Input
        double[] input = Genotype2.generateRandomGene();
        input[0] = FunctionLookup.getIndexForFunction("Input");
        input[1] = 100.0;
        input[2] = 100.0;
        input[6] = 0.0; // (output)
        genes[genoSize - 2] = input;

        // -Flop Top-
        // -> Change Connection
        double[] chc1 = Genotype2.generateRandomGene();
        chc1[0] = FunctionLookup.getIndexForFunction("SMchc");
        chc1[1] = 1.0;                          // connection 1
        chc1[2] = 2.0;                          // connection 2
        chc1[3] = 0.0;                          // real 1 (node to change)
        chc1[4] = 2.0;                          // real 2 (real 2 % 2 -> connection to change)
        chc1[5] = 5.0;                          // real 3 (new connection)
        chc1[6] = 0.0;                          // (output)
        genes[genoSize - 3] = chc1;

        // -> Input
        double[] input1 = Genotype2.generateRandomGene();
        input1[0] = FunctionLookup.getIndexForFunction("Input");
        genes[genoSize - 4] = input1;
        // -> Input Decrement
        double[] inputdecrement1 = Genotype2.generateRandomGene();
        inputdecrement1[0] = FunctionLookup.getIndexForFunction("Input Decrement");
        genes[genoSize - 5] = inputdecrement1;

        // -Flop Bottom-
        // -> Change Connection
        double[] chc2 = Genotype2.generateRandomGene();
        chc2[0] = FunctionLookup.getIndexForFunction("SMchc");
        chc2[1] = 1.0;                          // connection 1
        chc2[2] = 2.0;                          // connection 2
        chc2[3] = 0.0;                          // real 1 (node to change)
        chc2[4] = 2.0;                          // real 2 (real 2 % 2 -> connection to change)
        chc2[5] = 2.0;                          // real 3 (new connection)
        chc2[6] = 0.0;                          // (output)
        genes[genoSize - 6] = chc2;
        // -> Input
        double[] input2 = Genotype2.generateRandomGene();
        input2[0] = FunctionLookup.getIndexForFunction("Input");
        genes[genoSize - 7] = input2;
        // -> Input Decrement
        double[] inputdecrement2 = Genotype2.generateRandomGene();
        inputdecrement2[0] = FunctionLookup.getIndexForFunction("Input Decrement");
        genes[genoSize - 8] = inputdecrement2;

        // Fill up the rest - set all output false
        for (int i = 0; i < genoSize-8; i++) {
            genes[i] = Genotype2.generateRandomGene();
            genes[i][6] = 0.0;
        }

        return new Genotype2(genes);
    }

    public static Genotype2 getGenotype () {
        if (current == null) {
            current = connectionFloppingGenotype();
        }
        return current;
    }

    public static void mutate() {
        current.mutate(0.05, true);
    }

    public static void testPrint() {
        Genotype2 geno = connectionFloppingGenotype();
        Phenotype2 pheno = new Phenotype2(geno);
        Candidate cand = new Candidate(pheno);

        // Print node functions
        double[][] nodes = geno.copyGenes();
        int ocunt = 0;
        for (double[] node : nodes) {
            System.out.println(ocunt + " - " + node[0]);
            System.out.println(FunctionLookup.getFunction((int) node[0]));
            ocunt++;
        }
        //System.out.println(Arrays.deepToString(geno.copyGenes()));
        //System.out.println(cand.genotypeTree(0));
        System.out.println(cand.toStringWithNodeTree());
    }
}
