package no.ntnu.cerebrum.medulla.smcgp.graph.action;

import org.slf4j.Logger;

public abstract class SMFunction<T> {

    // Logging
    private static Logger log = org.slf4j.LoggerFactory.getLogger(SMFunction.class);
    protected void logging(Object cls, String msg) {
        log.debug(cls.getClass().getSimpleName() + " " + msg);
    }

    public abstract T doSMFunction(double[][] graph, double[] activatedNode, int position);

    /** Helper method for getting real sm parameters
     *
     * @param currentNode the node that is activated
     * @param p the index of the real value sm parameter [0-2]
     * @return the value at position p
     */
    public int getParameter(double[] currentNode, int p) {
        return (int) currentNode[p + 3];
    }

    /**
     *  Method to ensure that all index pointer are within the graph
     *
     * @param outBounds the unchecked position in the graph
     * @param graphSize the size of the graph
     * @return int position within bounds
     */
    public int inBounds(int outBounds, int graphSize) {
        return Math.min(Math.abs(outBounds), graphSize - 1);
    }

}
