package no.ntnu.cerebrum.medulla.smcgp.graph.action.selfmodifying;

import no.ntnu.cerebrum.medulla.config.Config;
import no.ntnu.cerebrum.medulla.smcgp.graph.action.SMFunction;

import java.util.Arrays;

public class SMdup extends SMFunction {
    @Override
    public Object doSMFunction(double[][] graph, double[] activatedNode, int position) {
        int p0 = getParameter(activatedNode, 0);
        int p1 = getParameter(activatedNode, 1);
        int p2 = getParameter(activatedNode, 2);

        int from = inBounds(p0 + position, graph.length);
        int to = inBounds(from + p1, graph.length);
        int insertAfter = inBounds(from + p2, graph.length);

        int maxMov = Config.getInt("no.ntnu.cerebrum.medulla.smcgp.maxNodesMovedOrDuplicated");
        int maxNodeSize = (int) (Config.getInt("no.ntnu.cerebrum.medulla.smcgp.startNodes") * Config.getDouble("no.ntnu.cerebrum.medulla.smcgp.maxGrowthFactor"));

        if (to > from) {
            // Place an upper limit to the amount of nodes that can be duplicated each iteration
            if (to - from > maxMov) to = from + maxMov;

            // If we are about to exceed the number of nodes set by max growth factor, just return the graph
            if (graph.length + (to - from) > maxNodeSize) return graph;

            // Copy duplicates
            double[][] duplicates = new double[to - from][];
            for (int dupIndex = 0; dupIndex < (to - from); dupIndex++) {
                duplicates[dupIndex] = Arrays.copyOf(graph[from + dupIndex], graph[from + dupIndex].length);
            }

            // Create new graph and insert at position @insertAfter
            double[][] newGraph = new double[graph.length + duplicates.length][];
            int nodesToBePlaced = to - from;
            for (int i = 0; i < newGraph.length; i++) {

                if (i <= insertAfter) {
                    newGraph[i] = graph[i];
                } else if (i > insertAfter && nodesToBePlaced > 0) {
                    newGraph[i] = duplicates[(to - from) - nodesToBePlaced];
                    nodesToBePlaced--;
                } else if (nodesToBePlaced == 0) {
                    newGraph[i] = graph[i - (to - from)];
                }

            }
            return newGraph;
        }

        return graph;
    }
}
