package no.ntnu.cerebrum.medulla.telemetry;

import com.google.gson.Gson;
import no.ntnu.cerebrum.medulla.Main;

import java.util.concurrent.ConcurrentHashMap;

/**
 * Created 2016-09-20.
 *
 * cerebrum-medulla is licenced under the MIT licence.
 */
public class Aggregator {
    private ConcurrentHashMap<String, Double> metrics;
    private Gson gson;

    /**
     * Constructs an Aggregator object.
     */
    public Aggregator() {
        metrics = new ConcurrentHashMap<>();
        gson = new Gson();

        metrics.put("errors", 0.0);
        metrics.put("warnings", 0.0);
    }

    /**
     * Serializes the contents of the metrics map as JSON
     * @return A JSON encoded string representation of the metrics map.
     */
    public String toJSON() {
        return gson.toJson(metrics);
    }

    /**
     * Serializes the contents of the metrics map to Prometheus metrics definitions
     * @return A Prometheus formatted representation of the metrics map.
     */
    public String toPrometheusFormat() {
        return "# HELP medulla_errors The total number of errors this run.\n" +
            "# TYPE medulla_errors counter\n" +
            String.format("medulla_errors{host=%s} %.0f\n", Main.hostname, metrics.get("errors")) +
            "# HELP medulla_warnings The total number of warnings this run.\n" +
            "# TYPE medulla_warnings counter\n" +
            String.format("medulla_warnings{host=%s} %.0f\n", Main.hostname, metrics.get("warnings"));
    }
}
