package no.ntnu.cerebrum.medulla.telemetry;

import no.ntnu.cerebrum.medulla.smcgp.evolution.EvolutionListener;
import org.apache.commons.math3.stat.descriptive.SummaryStatistics;
import static no.ntnu.cerebrum.medulla.smcgp.evolution.EvolutionListener.KillReason.*;

import java.util.HashMap;

/**
 * Created 2017-05-05.
 * <p>
 * cerebrum-medulla is licenced under the MIT licence.
 */
public class EvolveStatsCollector implements EvolutionListener {
    private SummaryStatistics fitnessStats;
    private SummaryStatistics numIterationsStats;

    private HashMap<KillReason, Integer> killReasons;

    public EvolveStatsCollector() {
        fitnessStats = new SummaryStatistics();
        numIterationsStats = new SummaryStatistics();
        killReasons = new HashMap<>();
    }

    @Override
    public void fitnessEvaluationComplete(double fitness, int numIterations, KillReason reason) {
        killReasons.put(reason, killReasons.getOrDefault(reason, 0) + 1);
        fitnessStats.addValue(fitness);
        numIterationsStats.addValue(numIterations);
    }

    public String toString() {
        return String.format(
            "{N:%d MEAF:%.3f IT:%.3f/%.0f, KR: al:%d/c:%d/st:%d/inSm:%d/inp:%d/n:%d/sh:%d/t:%d/uv:%d}",
            fitnessStats.getN(),
            fitnessStats.getMean(),
            numIterationsStats.getMean(),
            numIterationsStats.getSum(),
            killReasons.getOrDefault(ATTRACTOR_LENGTH, 0),
            killReasons.getOrDefault(CONVERGED, 0),
            killReasons.getOrDefault(FEW_STATES_OR_STABLE, 0),
            killReasons.getOrDefault(INPUT_OR_SM, 0),
            killReasons.getOrDefault(INPUT_PERCENTAGE, 0),
            killReasons.getOrDefault(NONE, 0),
            killReasons.getOrDefault(SPECTRUM_HARMONICS, 0),
            killReasons.getOrDefault(TIMEOUT, 0),
            killReasons.getOrDefault(UNIQUE_VALUES, 0)
        );
    }
}
